@@ -4,10 +4,14 @@ This module is a collection of numba extensions that perform operations
 of varying levels of danger and complexity needed to make this monstrosity
 work correctly.
 """
+
+import contextlib
 import inspect
-from typing import Callable, Optional, Tuple, Union
+import operator
+from typing import Callable, Generator, Optional, Tuple
 
 import numba
+from llvmlite import ir
 from llvmlite.ir.instructions import (
     CallInstr,
     CastInstr,
@@ -19,16 +23,16 @@ from llvmlite.ir.instructions import (
 )
 from llvmlite.llvmpy.core import Builder
 from numba import extending, types
-from numba.core import cgutils, imputils
+from numba.core import cgutils, imputils, pythonapi
 from numba.core.base import BaseContext
 from numba.core.typing import ctypes_utils
 from numba.core.typing.context import Context
 from numba.core.typing.templates import Signature
 
 from .sqlite import (
-    RESULT_SETTERS,
+    SQLITE3_RESULT_SETTERS,
+    SQLITE3_VALUE_EXTRACTORS,
     SQLITE_NULL,
-    VALUE_EXTRACTORS,
     sqlite3_value_type,
     strlen,
 )
@@ -36,6 +40,7 @@ from .sqlite import (
 
 def _add_linking_libs(context: BaseContext, call: CallInstr) -> None:
     """Add the required libs for the callable to allow inlining."""
+
     try:
         libs = call.libs
     except AttributeError:
@@ -44,15 +49,26 @@ def _add_linking_libs(context: BaseContext, call: CallInstr) -> None:
         context.add_linking_libs(libs)
 
 
+@contextlib.contextmanager
+def gil(pyapi: pythonapi.PythonAPI) -> Generator[ir.AllocaInstr, None, None]:
+    """Generate instructions to hold the GIL."""
+    locked_gil = pyapi.gil_ensure()
+    try:
+        yield locked_gil
+    finally:
+        pyapi.gil_release(locked_gil)
+
+
 @extending.intrinsic  # type: ignore[misc]
 def unsafe_cast(
     typingctx: Context,
-    src: Union[types.RawPointer, types.Integer],
+    src: types.Integer,
     dst: types.ClassType,
 ) -> Tuple[
-    Signature, Callable[[BaseContext, Builder, Signature, Tuple[Value, ...]], LoadInstr]
+    Signature,
+    Callable[[BaseContext, Builder, Signature, Tuple[Value, Value]], LoadInstr],
 ]:
-    """Cast a void pointer to a jitclass.
+    """Cast a void pointer to a `jitclass` type.
 
     Parameters
     ----------
@@ -68,9 +84,7 @@ def unsafe_cast(
         If `src` is not a raw pointer or integer or `dst` is not a jitclass
         type.
     """
-    if isinstance(src, (types.RawPointer, types.Integer)) and isinstance(
-        dst, types.ClassType
-    ):
+    if isinstance(src, types.Integer) and isinstance(dst, types.ClassType):
         inst_typ = dst.instance_type
         sig = inst_typ(types.voidptr, dst)
 
@@ -78,7 +92,7 @@ def unsafe_cast(
             context: BaseContext,
             builder: Builder,
             signature: Signature,
-            args: Tuple[Value, ...],
+            args: Tuple[Value, Value],
         ) -> LoadInstr:
             ptr, _ = args
 
@@ -116,9 +130,10 @@ def init(
     inst_typ: types.ClassInstanceType,
     user_data: types.Integer,
 ) -> Tuple[
-    Signature, Callable[[BaseContext, Builder, Signature, Tuple[Value, ...]], None]
+    Signature,
+    Callable[[BaseContext, Builder, Signature, Tuple[Value, Value]], None],
 ]:
-    """Initialize a jitclass by calling its constructor.
+    """Initialize a `jitclass` by calling its constructor.
 
     Parameters
     ----------
@@ -133,23 +148,22 @@ def init(
             context: BaseContext,
             builder: Builder,
             signature: Signature,
-            args: Tuple[Value, ...],
+            args: Tuple[Value, Value],
         ) -> None:
             instance, user_data = args
 
             # cast the user-defined void* payload to bool*
             raw = builder.bitcast(user_data, cgutils.bool_t.as_pointer())
 
-            # generate a basic block to check if the constructor has been
-            # called
+            # generate an if statement to check whether the constructor has
+            # been called
             #
-            # it's only ever called once, so set likely=False for better cache
+            # it's only ever called once, so use `if_unlikely` for better
             # locality
-            with builder.if_then(builder.not_(builder.load(raw)), likely=False):
+            with cgutils.if_unlikely(builder, builder.not_(builder.load(raw))):
                 # pull out the function pointer
                 dist_typ = types.Dispatcher(inst_typ.jit_methods["__init__"])
-                fnty = types.void(inst_typ)
-                fn = context.get_function(dist_typ, fnty)
+                fn = context.get_function(dist_typ, types.void(inst_typ))
 
                 _add_linking_libs(context, fn)
 
@@ -182,7 +196,10 @@ def python_signature_to_numba_signature(
             extending.as_numba_type(self_type if first_name == "self" else first_type)
         )
 
-    return_type = extending.as_numba_type(signature.return_annotation or types.void)
+    return_ann = signature.return_annotation
+    return_type = extending.as_numba_type(
+        return_ann if return_ann is not None else types.void
+    )
 
     input_types.extend(
         extending.as_numba_type(param.annotation) for _, param in parameters
@@ -193,73 +210,66 @@ def python_signature_to_numba_signature(
 @extending.intrinsic  # type: ignore[misc]
 def reset_init(
     typingctx: Context,
-    user_data_type: types.Integer,
-) -> Tuple[
-    Signature, Callable[[BaseContext, Builder, Signature, Tuple[Value, ...]], None]
-]:
-    """Reset the init when finalize is called to ensure the constructor is called again.
+    user_data: types.Integer,
+) -> Tuple[Signature, Callable[[BaseContext, Builder, Signature, Tuple[Value]], None]]:
+    """Reset the user data when `finalize` is called.
+
+    This call ensures that the constructor is called the next time the `step`
+    method is invoked.
 
-    Parameters
-    ----------
-    typingctx
-    user_data_type
     """
-    if isinstance(user_data_type, types.Integer):
+    if isinstance(user_data, types.Integer):
         sig = types.void(types.voidptr)
 
         def codegen(
             context: BaseContext,
             builder: Builder,
             signature: Signature,
-            args: Tuple[Value, ...],
+            args: Tuple[Value],
         ) -> None:
             (user_data,) = args
             # cast the user defined data structure, (which is current a pointer
             # to a boolean indicating whether the constructor has been called)
             # to a pointer to bool
             #
-            # we pass around voidptrs because that's really the only way to
-            # customize end-user data structures in C
+            # void pointer is the typical way to customize end-user data
+            # structures in C
             raw = builder.bitcast(user_data, cgutils.bool_t.as_pointer())
             builder.store(context.get_constant(types.boolean, False), raw)
 
         return sig, codegen
-    raise TypeError(f"Unable to uninitialize type {user_data_type}")
+
+    raise TypeError(f"Unable to uninitialize flag of type `{user_data}`")
 
 
 @extending.intrinsic  # type: ignore[misc]
 def safe_decref(
-    typingctx: Context, pyobject: types.RawPointer
-) -> Tuple[
-    Signature, Callable[[BaseContext, Builder, Signature, Tuple[Value, ...]], None]
-]:
-    """Safely decrement the reference count of a PyObject*.
-
-    Parameters
-    ----------
-    typingctx
-    user_data_type
-    """
-    if isinstance(pyobject, types.RawPointer):
+    typingctx: Context,
+    pyobject_type: types.RawPointer,
+) -> Tuple[Signature, Callable[[BaseContext, Builder, Signature, Tuple[Value]], None]]:
+    """Safely decrement the reference count of a Python object."""
+    if isinstance(pyobject_type, types.RawPointer):
         sig = types.void(types.voidptr)
 
         def codegen(
             context: BaseContext,
             builder: Builder,
             signature: Signature,
-            args: Tuple[Value, ...],
+            args: Tuple[Value],
         ) -> None:
             (user_data,) = args
             pyapi = context.get_python_api(builder)
             # check whether the pyobject is null, (likely not)
-            with builder.if_then(cgutils.is_not_null(builder, user_data), likely=True):
+            with cgutils.if_likely(builder, cgutils.is_not_null(builder, user_data)):
                 # we can't call decref safely without holding the GIL
-                gil = pyapi.gil_ensure()
-                pyapi.decref(user_data)
-                pyapi.gil_release(gil)
+                with gil(pyapi):
+                    pyapi.decref(user_data)
 
         return sig, codegen
-    raise TypeError(f"Unable to decref type {pyobject}")
+
+    raise TypeError(
+        f"Unable to decrement the reference count of type `{pyobject_type}`"
+    )
 
 
 @extending.intrinsic  # type: ignore[misc]
@@ -267,7 +277,10 @@ def make_arg_tuple(
     typingctx: Context, func: types.Callable, argv: types.CPointer
 ) -> Tuple[
     Signature,
-    Callable[[BaseContext, Builder, Signature, Tuple[Value, ...]], InsertValue],
+    Callable[
+        [BaseContext, Builder, Signature, Tuple[Value, Value]],
+        InsertValue,
+    ],
 ]:
     """Construct a typed argument tuple to pass to a UD(A)F."""
     (func_type,), _ = func.get_call_signatures()
@@ -283,41 +296,48 @@ def make_arg_tuple(
         context: BaseContext,
         builder: Builder,
         signature: Signature,
-        args: Tuple[Value, ...],
+        args: Tuple[Value, Value],
     ) -> InsertValue:
         # first argument is the instance, and we don't need it here
         _, argv = args
+
+        # initialize a list to hold the converted function arguments
         converted_args = []
+
+        # grab the python API object, which we use in the case of encountering
+        # asn unexpected null value
         pyapi = context.get_python_api(builder)
 
         # make the SQLITE_NULL value type constant available
         sqlite_null = context.get_constant(types.int32, SQLITE_NULL)
 
-        for i, argtype in enumerate(argtypes):
-            # get a pointer to the sqlite3_value_type C function
-            sqlite3_value_type_numba = context.get_constant_generic(
-                builder,
-                ctypes_utils.make_function_type(sqlite3_value_type),
-                sqlite3_value_type,
-            )
+        # get a pointer to the sqlite3_value_type C function
+        sqlite3_value_type_numba = context.get_constant_generic(
+            builder,
+            ctypes_utils.make_function_type(sqlite3_value_type),
+            sqlite3_value_type,
+        )
 
+        for i, argtype in enumerate(argtypes):
             # get a pointer to the ith argument
-            element_pointer = cgutils.gep(builder, argv, i, inbounds=True)
+            sqlite3_value_pointer = cgutils.gep(builder, argv, i, inbounds=True)
 
             # deref that pointer
-            element = builder.load(element_pointer)
+            sqlite3_value = builder.load(sqlite3_value_pointer)
 
-            # get the value type
-            value_type = builder.call(sqlite3_value_type_numba, [element])
+            # the previous two instructions are equivalent to the following C code:
+            #
+            # sqlite3_value** args; // this is passed in
+            # args[i] // or *(args + i)
+
+            # call the SQLite C API to get the value type
+            value_type = builder.call(sqlite3_value_type_numba, [sqlite3_value])
 
             # check whether the value is equal to SQLITE_NULL
-            is_not_sqlite_null = cgutils.is_true(
-                builder, builder.icmp_signed("!=", value_type, sqlite_null)
-            )
+            is_not_sqlite_null = builder.icmp_signed("!=", value_type, sqlite_null)
 
-            # setup value extraction #
             # get the appropriate ctypes extraction routine
-            ctypes_function = VALUE_EXTRACTORS[argtype]
+            ctypes_function = SQLITE3_VALUE_EXTRACTORS[argtype]
 
             # create a numba function type for the converter
             converter = ctypes_utils.make_function_type(ctypes_function)
@@ -329,13 +349,13 @@ def make_arg_tuple(
             # type and make an optional value with it
             #
             # otherwise the raw value is the argument
-            raw = builder.call(fn, [element])
+            raw = builder.call(fn, [sqlite3_value])
             out_type = context.get_value_type(argtype)
             instr = cgutils.alloca_once(builder, out_type)
             underlying_type = getattr(argtype, "type", argtype)
 
             if isinstance(argtype, types.Optional):
-
+                # branch to handle null values
                 with builder.if_else(is_not_sqlite_null) as (then, otherwise):
                     with then:
                         # you _must_ put code that only executes in this block,
@@ -354,7 +374,7 @@ def make_arg_tuple(
                                 else map_sqlite_string_to_numba_uni_str(
                                     context,
                                     builder,
-                                    data=builder.inttoptr(raw, element.type),
+                                    data=builder.inttoptr(raw, sqlite3_value.type),
                                 )
                             ),
                         )
@@ -365,8 +385,8 @@ def make_arg_tuple(
                         none = context.make_optional_none(builder, underlying_type)
                         builder.store(none, instr)
             else:
-                # check for NULLs, and raise an exception if the value is NULL,
-                # because the input doesn't have an option type
+                # raise an exception if the value is NULL, because the input
+                # type is not optional and therefore cannot handle NULLs
                 #
                 # favor the branch where the value isn't null, since it's
                 # an error condition to accept null values without an option type
@@ -381,7 +401,7 @@ def make_arg_tuple(
                             else map_sqlite_string_to_numba_uni_str(
                                 context,
                                 builder,
-                                data=builder.inttoptr(raw, element.type),
+                                data=builder.inttoptr(raw, sqlite3_value.type),
                             )
                         )
 
@@ -390,24 +410,23 @@ def make_arg_tuple(
                     with otherwise:
                         # without the GIL here we're deep in undefined behavior
                         # land
-                        gil = pyapi.gil_ensure()
-                        pyapi.err_set_string(
-                            "PyExc_ValueError",
-                            (
-                                "encountered unexpected NULL in call to "
-                                "user-defined numba function "
-                                f"{func.dispatcher.py_func.__name__!r}"
-                            ),
-                        )
-                        pyapi.gil_release(gil)
+                        with gil(pyapi):
+                            pyapi.err_set_string(
+                                "PyExc_ValueError",
+                                (
+                                    "encountered unexpected NULL in call to "
+                                    "user-defined numba function "
+                                    f"{func.dispatcher.py_func.__name__!r}"
+                                ),
+                            )
 
             # instr is a pointer, so we need to dereference it to use it later
             # in the argument tuple
             converted_args.append(builder.load(instr))
 
         # construct a tuple of arguments (fixed length and known types)
-        res = context.make_tuple(builder, tuple_type, converted_args)
-        return imputils.impl_ret_new_ref(context, builder, tuple_type, res)
+        arg_tuple = context.make_tuple(builder, tuple_type, converted_args)
+        return imputils.impl_ret_untracked(context, builder, tuple_type, arg_tuple)
 
     return sig, codegen
 
@@ -475,31 +494,27 @@ def get_sqlite3_result_function(
     typingctx: Context, value_type: types.Type
 ) -> Tuple[
     Signature,
-    Callable[[BaseContext, Builder, Signature, Tuple[Value, ...]], CastInstr],
+    Callable[[BaseContext, Builder, Signature, Tuple[Value]], CastInstr],
 ]:
-    """Return the correct result setting function for the given type."""
-    underlying_type = getattr(value_type, "type", value_type)
-    func_type = types.void(types.voidptr, underlying_type)
+    """Return the correct result setting function (pointer) for the given type."""
+    input_type = getattr(value_type, "type", value_type)
+    func_type = types.void(types.voidptr, input_type)
 
     external_function_pointer = types.ExternalFunctionPointer(
-        func_type, ctypes_utils.get_pointer
+        func_type,
+        operator.attrgetter("address"),
     )
-    sig = external_function_pointer(underlying_type)
+    sig = external_function_pointer(input_type)
 
     def codegen(
         context: BaseContext,
         builder: Builder,
         signature: Signature,
-        args: Tuple[Value, ...],
+        args: Tuple[Value],
     ) -> CastInstr:
-        # get the appropriate setter function
-        result_setter = RESULT_SETTERS[value_type]
-
-        # create a numba function type for the converter
-        converter = ctypes_utils.make_function_type(result_setter)
-
-        # get the function pointer instruction out
-        return context.get_constant_generic(builder, converter, result_setter)
+        return context.get_constant_generic(
+            builder, external_function_pointer, SQLITE3_RESULT_SETTERS[value_type]
+        )
 
     return sig, codegen
 
@@ -509,7 +524,7 @@ def sizeof(
     typingctx: Context, src: types.ClassType
 ) -> Tuple[
     Signature,
-    Callable[[BaseContext, Builder, Signature, Tuple[Value, ...]], Constant],
+    Callable[[BaseContext, Builder, Signature, Tuple[Value]], Constant],
 ]:
     """Return the size in bytes of a type."""
     if isinstance(src, types.ClassType):
@@ -519,15 +534,15 @@ def sizeof(
             context: BaseContext,
             builder: Builder,
             signature: Signature,
-            args: Tuple[Value, ...],
+            args: Tuple[Value],
         ) -> Constant:
             data_type = context.get_data_type(src.instance_type)
             size_of_data_type = context.get_abi_sizeof(data_type)
             return context.get_constant(sig.return_type, size_of_data_type)
 
         return sig, codegen
-    else:
-        raise TypeError("Cannot get sizeof non jitclass")
+
+    raise TypeError(f"Cannot get ABI size of `{src}`")
 
 
 @extending.intrinsic  # type: ignore[misc]
@@ -550,4 +565,7 @@ def is_not_null_pointer(
 
         return sig, codegen
 
-    raise TypeError(f"Cannot check whether a {raw_pointer_type} is not a null pointer")
+    raise TypeError(
+        f"Cannot check whether a value of type `{raw_pointer_type}` "
+        "is not a null pointer"
+    )
